@@ -112,6 +112,7 @@ location_memsize(const void *ptr)
 static const rb_data_type_t location_data_type = {
     "frame_info",
     {location_mark, location_free, location_memsize,},
+    NULL, NULL, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
 static inline rb_backtrace_location_t *
@@ -142,6 +143,14 @@ location_lineno(rb_backtrace_location_t *loc)
     }
 }
 
+/*
+ * Returns the line number of this frame.
+ *
+ * For example, using +caller_locations.rb+ from Thread::Backtrace::Location
+ *
+ *	loc = c(0..1).first
+ *	loc.lineno #=> 2
+ */
 static VALUE
 location_lineno_m(VALUE self)
 {
@@ -164,6 +173,33 @@ location_label(rb_backtrace_location_t *loc)
     }
 }
 
+/*
+ * Returns the label of this frame.
+ *
+ * Usually consists of method, class, module, etc names with decoration.
+ *
+ * Consider the following example:
+ *
+ *	def foo
+ *	  puts caller_locations(0).first.label
+ *
+ *	  1.times do
+ *	    puts caller_locations(0).first.label
+ *
+ *	    1.times do
+ *	      puts caller_locations(0).first.label
+ *	    end
+ *
+ *	  end
+ *	end
+ *
+ * The result of calling +foo+ is this:
+ *
+ *	label: foo
+ *	label: block in foo
+ *	label: block (2 levels) in foo
+ *
+ */
 static VALUE
 location_label_m(VALUE self)
 {
@@ -186,6 +222,11 @@ location_base_label(rb_backtrace_location_t *loc)
     }
 }
 
+/*
+ * Returns the base label of this frame.
+ *
+ * Usually same as #label, without decoration.
+ */
 static VALUE
 location_base_label_m(VALUE self)
 {
@@ -211,6 +252,14 @@ location_path(rb_backtrace_location_t *loc)
     }
 }
 
+/*
+ * Returns the file name of this frame.
+ *
+ * For example, using +caller_locations.rb+ from Thread::Backtrace::Location
+ *
+ *	loc = c(0..1).first
+ *	loc.path #=> caller_locations.rb
+ */
 static VALUE
 location_path_m(VALUE self)
 {
@@ -236,6 +285,11 @@ location_absolute_path(rb_backtrace_location_t *loc)
     }
 }
 
+/*
+ * Returns the full file path of this frame.
+ *
+ * Same as #path, but includes the absolute path.
+ */
 static VALUE
 location_absolute_path_m(VALUE self)
 {
@@ -294,12 +348,19 @@ location_to_str(rb_backtrace_location_t *loc)
     return location_format(file, lineno, name);
 }
 
+/*
+ * Returns a Kernel#caller style string representing this frame.
+ */
 static VALUE
 location_to_str_m(VALUE self)
 {
     return location_to_str(location_ptr(self));
 }
 
+/*
+ * Returns the same as calling +inspect+ on the string representation of
+ * #to_str
+ */
 static VALUE
 location_inspect_m(VALUE self)
 {
@@ -347,6 +408,7 @@ backtrace_memsize(const void *ptr)
 static const rb_data_type_t backtrace_data_type = {
     "backtrace",
     {backtrace_mark, backtrace_free, backtrace_memsize,},
+    NULL, NULL, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
 int
@@ -586,13 +648,13 @@ backtrace_load_data(VALUE self, VALUE str)
 }
 
 VALUE
-vm_backtrace_str_ary(rb_thread_t *th, long lev, long n)
+rb_vm_backtrace_str_ary(rb_thread_t *th, long lev, long n)
 {
     return backtrace_to_str_ary(backtrace_object(th), lev, n);
 }
 
 VALUE
-vm_backtrace_location_ary(rb_thread_t *th, long lev, long n)
+rb_vm_backtrace_location_ary(rb_thread_t *th, long lev, long n)
 {
     return backtrace_to_location_ary(backtrace_object(th), lev, n);
 }
@@ -708,10 +770,39 @@ rb_backtrace(void)
     vm_backtrace_print(stderr);
 }
 
+static void
+oldbt_print_to(void *data, VALUE file, int lineno, VALUE name)
+{
+    VALUE output = (VALUE)data;
+    VALUE str = rb_sprintf("\tfrom %"PRIsVALUE":%d:in ", file, lineno);
+
+    if (NIL_P(name)) {
+	rb_str_cat2(str, "unknown method\n");
+    }
+    else {
+	rb_str_catf(str, " `%"PRIsVALUE"'\n", name);
+    }
+    rb_io_write(output, str);
+}
+
+void
+rb_backtrace_print_to(VALUE output)
+{
+    struct oldbt_arg arg;
+
+    arg.func = oldbt_print_to;
+    arg.data = (void *)output;
+    backtrace_each(GET_THREAD(),
+		   oldbt_init,
+		   oldbt_iter_iseq,
+		   oldbt_iter_cfunc,
+		   &arg);
+}
+
 VALUE
 rb_make_backtrace(void)
 {
-    return vm_backtrace_str_ary(GET_THREAD(), 0, 0);
+    return rb_vm_backtrace_str_ary(GET_THREAD(), 0, 0);
 }
 
 static VALUE
@@ -727,6 +818,8 @@ vm_backtrace_to_ary(rb_thread_t *th, int argc, VALUE *argv, int lev_default, int
 
     rb_scan_args(argc, argv, "02", &level, &vn);
 
+    if (argc == 2 && NIL_P(vn)) argc--;
+
     switch (argc) {
       case 0:
 	lev = lev_default + lev_plus;
@@ -755,11 +848,14 @@ vm_backtrace_to_ary(rb_thread_t *th, int argc, VALUE *argv, int lev_default, int
 	}
       case 2:
 	lev = NUM2LONG(level);
+	n = NUM2LONG(vn);
 	if (lev < 0) {
 	    rb_raise(rb_eArgError, "negative level (%ld)", lev);
 	}
+	if (n < 0) {
+	    rb_raise(rb_eArgError, "negative size (%ld)", n);
+	}
 	lev += lev_plus;
-	n = NUM2LONG(vn);
 	break;
       default:
 	lev = n = 0; /* to avoid warning */
@@ -793,30 +889,38 @@ thread_backtrace_to_ary(int argc, VALUE *argv, VALUE thval, int to_str)
 }
 
 VALUE
-vm_thread_backtrace(int argc, VALUE *argv, VALUE thval)
+rb_vm_thread_backtrace(int argc, VALUE *argv, VALUE thval)
 {
     return thread_backtrace_to_ary(argc, argv, thval, 1);
 }
 
 VALUE
-vm_thread_backtrace_locations(int argc, VALUE *argv, VALUE thval)
+rb_vm_thread_backtrace_locations(int argc, VALUE *argv, VALUE thval)
 {
     return thread_backtrace_to_ary(argc, argv, thval, 0);
 }
 
 /*
  *  call-seq:
- *     caller(start=1)    -> array or nil
+ *     caller(start=1, length=nil)  -> array or nil
+ *     caller(range)		    -> array or nil
  *
  *  Returns the current execution stack---an array containing strings in
- *  the form ``<em>file:line</em>'' or ``<em>file:line: in
- *  `method'</em>''. The optional _start_ parameter
- *  determines the number of initial stack entries to omit from the
- *  result.
+ *  the form <code>file:line</code> or <code>file:line: in
+ *  `method'</code>.
+ *
+ *  The optional _start_ parameter determines the number of initial stack
+ *  entries to omit from the top of the stack.
+ *
+ *  A second optional +length+ parameter can be used to limit how many entries
+ *  are returned from the stack.
  *
  *  Returns +nil+ if _start_ is greater than the size of
  *  current execution stack.
  *
+ *  Optionally you can pass a range, which will return an array containing the
+ *  entries within the specified range.
+ *
  *     def a(skip)
  *       caller(skip)
  *     end
@@ -840,6 +944,28 @@ rb_f_caller(int argc, VALUE *argv)
     return vm_backtrace_to_ary(GET_THREAD(), argc, argv, 1, 1, 1);
 }
 
+/*
+ *  call-seq:
+ *     caller_locations(start=1, length=nil)	-> array or nil
+ *     caller_locations(range)			-> array or nil
+ *
+ *  Returns the current execution stack---an array containing
+ *  backtrace location objects.
+ *
+ *  See Thread::Backtrace::Location for more information.
+ *
+ *  The optional _start_ parameter determines the number of initial stack
+ *  entries to omit from the top of the stack.
+ *
+ *  A second optional +length+ parameter can be used to limit how many entries
+ *  are returned from the stack.
+ *
+ *  Returns +nil+ if _start_ is greater than the size of
+ *  current execution stack.
+ *
+ *  Optionally you can pass a range, which will return an array containing the
+ *  entries within the specified range.
+ */
 static VALUE
 rb_f_caller_locations(int argc, VALUE *argv)
 {
@@ -850,13 +976,59 @@ rb_f_caller_locations(int argc, VALUE *argv)
 void
 Init_vm_backtrace(void)
 {
-    /* ::RubyVM::Backtrace */
+    /* :nodoc: */
     rb_cBacktrace = rb_define_class_under(rb_cThread, "Backtrace", rb_cObject);
     rb_define_alloc_func(rb_cBacktrace, backtrace_alloc);
     rb_undef_method(CLASS_OF(rb_cBacktrace), "new");
     rb_marshal_define_compat(rb_cBacktrace, rb_cArray, backtrace_dump_data, backtrace_load_data);
 
-    /* ::RubyVM::Backtrace::Location */
+    /*
+     *	An object representation of a stack frame, initialized by
+     *	Kernel#caller_locations.
+     *
+     *	For example:
+     *
+     *		# caller_locations.rb
+     *		def a(skip)
+     *		  caller_locations(skip)
+     *		end
+     *		def b(skip)
+     *		  a(skip)
+     *		end
+     *		def c(skip)
+     *		  b(skip)
+     *		end
+     *
+     *		c(0..2).map do |call|
+     *		  puts call.to_s
+     *		end
+     *
+     *	Running <code>ruby caller_locations.rb</code> will produce:
+     *
+     *		caller_locations.rb:2:in `a'
+     *		caller_locations.rb:5:in `b'
+     *		caller_locations.rb:8:in `c'
+     *
+     *	Here's another example with a slightly different result:
+     *
+     *		# foo.rb
+     *		class Foo
+     *		  attr_accessor :locations
+     *		  def initialize(skip)
+     *		    @locations = caller_locations(skip)
+     *		  end
+     *		end
+     *
+     *		Foo.new(0..2).locations.map do |call|
+     *		  puts call.to_s
+     *		end
+     *
+     *	Now run <code>ruby foo.rb</code> and you should see:
+     *
+     *		init.rb:4:in `initialize'
+     *		init.rb:8:in `new'
+     *		init.rb:8:in `<main>'
+     */
     rb_cBacktraceLocation = rb_define_class_under(rb_cBacktrace, "Location", rb_cObject);
     rb_undef_alloc_func(rb_cBacktraceLocation);
     rb_undef_method(CLASS_OF(rb_cBacktraceLocation), "new");
@@ -874,24 +1046,27 @@ Init_vm_backtrace(void)
 
 /* debugger API */
 
-#if defined __GNUC__ && __GNUC__ >= 4
-#pragma GCC visibility push(default)
-#endif
+RUBY_SYMBOL_EXPORT_BEGIN
 
-#if defined __GNUC__ && __GNUC__ >= 4
-#pragma GCC visibility pop
-#endif
+RUBY_SYMBOL_EXPORT_END
 
 struct rb_debug_inspector_struct {
     rb_thread_t *th;
     rb_control_frame_t *cfp;
     VALUE backtrace;
     VALUE contexts; /* [[klass, binding, iseq, cfp], ...] */
-    int backtrace_size;
+    long backtrace_size;
+};
+
+enum {
+    CALLER_BINDING_SELF,
+    CALLER_BINDING_CLASS,
+    CALLER_BINDING_BINDING,
+    CALLER_BINDING_ISEQ,
+    CALLER_BINDING_CFP
 };
 
 struct collect_caller_bindings_data {
-    rb_thread_t *th;
     VALUE ary;
 };
 
@@ -901,37 +1076,82 @@ collect_caller_bindings_init(void *arg, size_t size)
     /* */
 }
 
+static VALUE
+get_klass(const rb_control_frame_t *cfp)
+{
+    VALUE klass;
+    if (rb_vm_control_frame_id_and_class(cfp, 0, &klass)) {
+	if (RB_TYPE_P(klass, T_ICLASS)) {
+	    return RBASIC(klass)->klass;
+	}
+	else {
+	    return klass;
+	}
+    }
+    else {
+	return Qnil;
+    }
+}
+
 static void
 collect_caller_bindings_iseq(void *arg, const rb_control_frame_t *cfp)
 {
     struct collect_caller_bindings_data *data = (struct collect_caller_bindings_data *)arg;
-    rb_ary_push(data->ary,
-		rb_ary_new3(4,
-			    cfp->klass,
-			    rb_binding_new_with_cfp(data->th, cfp),
-			    cfp->iseq ? cfp->iseq->self : Qnil,
-			    GC_GUARDED_PTR(cfp)));
+    VALUE frame = rb_ary_new2(5);
+
+    rb_ary_store(frame, CALLER_BINDING_SELF, cfp->self);
+    rb_ary_store(frame, CALLER_BINDING_CLASS, get_klass(cfp));
+    rb_ary_store(frame, CALLER_BINDING_BINDING, GC_GUARDED_PTR(cfp)); /* create later */
+    rb_ary_store(frame, CALLER_BINDING_ISEQ, cfp->iseq ? cfp->iseq->self : Qnil);
+    rb_ary_store(frame, CALLER_BINDING_CFP, GC_GUARDED_PTR(cfp));
+
+    rb_ary_push(data->ary, frame);
 }
 
 static void
 collect_caller_bindings_cfunc(void *arg, const rb_control_frame_t *cfp, ID mid)
 {
     struct collect_caller_bindings_data *data = (struct collect_caller_bindings_data *)arg;
-    rb_ary_push(data->ary, rb_ary_new3(2, cfp->klass, Qnil));
+    VALUE frame = rb_ary_new2(5);
+
+    rb_ary_store(frame, CALLER_BINDING_SELF, cfp->self);
+    rb_ary_store(frame, CALLER_BINDING_CLASS, get_klass(cfp));
+    rb_ary_store(frame, CALLER_BINDING_BINDING, Qnil); /* not available */
+    rb_ary_store(frame, CALLER_BINDING_ISEQ, Qnil); /* not available */
+    rb_ary_store(frame, CALLER_BINDING_CFP, GC_GUARDED_PTR(cfp));
+
+    rb_ary_push(data->ary, frame);
 }
 
 static VALUE
 collect_caller_bindings(rb_thread_t *th)
 {
     struct collect_caller_bindings_data data;
+    VALUE result;
+    int i;
+
     data.ary = rb_ary_new();
-    data.th = th;
+
     backtrace_each(th,
 		   collect_caller_bindings_init,
 		   collect_caller_bindings_iseq,
 		   collect_caller_bindings_cfunc,
 		   &data);
-    return rb_ary_reverse(data.ary);
+
+    result = rb_ary_reverse(data.ary);
+
+    /* bindings should be created from top of frame */
+    for (i=0; i<RARRAY_LEN(result); i++) {
+	VALUE entry = rb_ary_entry(result, i);
+	VALUE cfp_val = rb_ary_entry(entry, CALLER_BINDING_BINDING);
+
+	if (!NIL_P(cfp_val)) {
+	    rb_control_frame_t *cfp = GC_GUARDED_PTR_REF(cfp_val);
+	    rb_ary_store(entry, CALLER_BINDING_BINDING, rb_binding_new_with_cfp(th, cfp));
+	}
+    }
+
+    return result;
 }
 
 /*
@@ -949,8 +1169,8 @@ rb_debug_inspector_open(rb_debug_inspector_func_t func, void *data)
 
     dbg_context.th = th;
     dbg_context.cfp = dbg_context.th->cfp;
-    dbg_context.backtrace = vm_backtrace_location_ary(th, 0, 0);
-    dbg_context.backtrace_size = RARRAY_LENINT(dbg_context.backtrace);
+    dbg_context.backtrace = rb_vm_backtrace_location_ary(th, 0, 0);
+    dbg_context.backtrace_size = RARRAY_LEN(dbg_context.backtrace);
     dbg_context.contexts = collect_caller_bindings(th);
 
     TH_PUSH_TAG(th);
@@ -969,7 +1189,7 @@ rb_debug_inspector_open(rb_debug_inspector_func_t func, void *data)
 }
 
 static VALUE
-frame_get(const rb_debug_inspector_t *dc, int index)
+frame_get(const rb_debug_inspector_t *dc, long index)
 {
     if (index < 0 || index >= dc->backtrace_size) {
 	rb_raise(rb_eArgError, "no such frame");
@@ -978,24 +1198,31 @@ frame_get(const rb_debug_inspector_t *dc, int index)
 }
 
 VALUE
-rb_debug_inspector_frame_class_get(const rb_debug_inspector_t *dc, int index)
+rb_debug_inspector_frame_self_get(const rb_debug_inspector_t *dc, long index)
 {
     VALUE frame = frame_get(dc, index);
-    return rb_ary_entry(frame, 0);
+    return rb_ary_entry(frame, CALLER_BINDING_SELF);
 }
 
 VALUE
-rb_debug_inspector_frame_binding_get(const rb_debug_inspector_t *dc, int index)
+rb_debug_inspector_frame_class_get(const rb_debug_inspector_t *dc, long index)
 {
     VALUE frame = frame_get(dc, index);
-    return rb_ary_entry(frame, 1);
+    return rb_ary_entry(frame, CALLER_BINDING_CLASS);
 }
 
 VALUE
-rb_debug_inspector_frame_iseq_get(const rb_debug_inspector_t *dc, int index)
+rb_debug_inspector_frame_binding_get(const rb_debug_inspector_t *dc, long index)
 {
     VALUE frame = frame_get(dc, index);
-    return rb_ary_entry(frame, 2);
+    return rb_ary_entry(frame, CALLER_BINDING_BINDING);
+}
+
+VALUE
+rb_debug_inspector_frame_iseq_get(const rb_debug_inspector_t *dc, long index)
+{
+    VALUE frame = frame_get(dc, index);
+    return rb_ary_entry(frame, CALLER_BINDING_ISEQ);
 }
 
 VALUE
@@ -1004,3 +1231,138 @@ rb_debug_inspector_backtrace_locations(const rb_debug_inspector_t *dc)
     return dc->backtrace;
 }
 
+int
+rb_profile_frames(int start, int limit, VALUE *buff, int *lines)
+{
+    int i;
+    rb_thread_t *th = GET_THREAD();
+    rb_control_frame_t *cfp = th->cfp, *end_cfp = RUBY_VM_END_CONTROL_FRAME(th);
+
+    for (i=0; i<limit && cfp != end_cfp;) {
+	if (cfp->iseq && cfp->pc) { /* should be NORMAL_ISEQ */
+	    if (start > 0) {
+		start--;
+		continue;
+	    }
+
+	    /* record frame info */
+	    buff[i] = cfp->iseq->self;
+	    if (lines) lines[i] = calc_lineno(cfp->iseq, cfp->pc);
+	    i++;
+	}
+	cfp = RUBY_VM_PREVIOUS_CONTROL_FRAME(cfp);
+    }
+
+    return i;
+}
+
+#define frame2iseq(frame) frame
+
+VALUE
+rb_profile_frame_path(VALUE frame)
+{
+    return rb_iseq_path(frame2iseq(frame));
+}
+
+VALUE
+rb_profile_frame_absolute_path(VALUE frame)
+{
+    return rb_iseq_absolute_path(frame2iseq(frame));
+}
+
+VALUE
+rb_profile_frame_label(VALUE frame)
+{
+    return rb_iseq_label(frame2iseq(frame));
+}
+
+VALUE
+rb_profile_frame_base_label(VALUE frame)
+{
+    return rb_iseq_base_label(frame2iseq(frame));
+}
+
+VALUE
+rb_profile_frame_first_lineno(VALUE frame)
+{
+    return rb_iseq_first_lineno(frame2iseq(frame));
+}
+
+VALUE
+rb_profile_frame_classpath(VALUE frame)
+{
+    VALUE klass = rb_iseq_klass(frame2iseq(frame));
+
+    if (klass && !NIL_P(klass)) {
+	if (RB_TYPE_P(klass, T_ICLASS)) {
+	    klass = RBASIC(klass)->klass;
+	}
+	else if (FL_TEST(klass, FL_SINGLETON)) {
+	    klass = rb_ivar_get(klass, id__attached__);
+	    if (!RB_TYPE_P(klass, T_CLASS))
+		return rb_sprintf("#<%s:%p>", rb_class2name(rb_obj_class(klass)), (void*)klass);
+	}
+	return rb_class_path(klass);
+    }
+    else {
+	return Qnil;
+    }
+}
+
+VALUE
+rb_profile_frame_singleton_method_p(VALUE frame)
+{
+    VALUE klass = rb_iseq_klass(frame2iseq(frame));
+    if (klass && !NIL_P(klass) && FL_TEST(klass, FL_SINGLETON)) {
+	return Qtrue;
+    }
+    else {
+	return Qfalse;
+    }
+}
+
+VALUE
+rb_profile_frame_method_name(VALUE frame)
+{
+    return rb_iseq_method_name(frame2iseq(frame));
+}
+
+VALUE
+rb_profile_frame_qualified_method_name(VALUE frame)
+{
+    VALUE method_name = rb_iseq_method_name(frame2iseq(frame));
+    if (method_name != Qnil) {
+	VALUE classpath = rb_profile_frame_classpath(frame);
+	VALUE singleton_p = rb_profile_frame_singleton_method_p(frame);
+
+	if (classpath != Qnil) {
+	    return rb_sprintf("%"PRIsVALUE"%s%"PRIsVALUE,
+			      classpath, singleton_p == Qtrue ? "." : "#", method_name);
+	}
+	else {
+	    return method_name;
+	}
+    }
+    else {
+	return Qnil;
+    }
+}
+
+VALUE
+rb_profile_frame_full_label(VALUE frame)
+{
+    VALUE label = rb_profile_frame_label(frame);
+    VALUE base_label = rb_profile_frame_base_label(frame);
+    VALUE qualified_method_name = rb_profile_frame_qualified_method_name(frame);
+
+    if (NIL_P(qualified_method_name) || base_label == qualified_method_name) {
+	return label;
+    }
+    else {
+	long label_length = RSTRING_LEN(label);
+	long base_label_length = RSTRING_LEN(base_label);
+	int prefix_len = rb_long2int(label_length - base_label_length);
+
+	return rb_sprintf("%.*s%"PRIsVALUE, prefix_len, RSTRING_PTR(label), qualified_method_name);
+    }
+}
